# -*- coding: utf-8 -*-

import simplejson

from django import forms
from django.conf import settings
from django.core.urlresolvers import reverse
from django.contrib import messages
from django.forms import ModelForm
from django.utils.translation import ugettext as _
from django.utils.safestring import mark_safe
from django_select2.fields import ModelSelect2MultipleField

from acomea.contact.models import Group
from acomea.mailup.models import Import
from acomea.mailup.tasks import ImportAndSyncContactsTask
from acomea.mailup.utils.client import MailUpClient


class WizardFileSelectionForm(ModelForm):

    class Meta:
        model = Import
        fields = ['file']


class WizardGroupsSelectionForm(ModelForm):
    groups = ModelSelect2MultipleField(
        label=_('Groups'), queryset=Group.objects, required=True
    )

    class Meta:
        model = Import
        fields = ['groups']


class WizardFieldsConfigurationForm(ModelForm):
    enabled_lists = [3, 4]

    class Meta:
        model = Import
        fields = ['mailup_status', 'change_db_status', 'skip_first_row']

    def __init__(self, request, *args, **kwargs):
        super(WizardFieldsConfigurationForm, self).__init__(*args, **kwargs)
        self.request = request
        mailup_client = MailUpClient()
        mailup_client.retrieve_access_token()
        mailup_lists = mailup_client.read_lists()
        self.fields['list'] = forms.ChoiceField(
            choices=[
                (int(item['idList']), item['Name']) \
                    for item in mailup_lists['Items'] if int(item['idList']) in self.enabled_lists
            ]
        )
        for col_pos in range(self.instance.file_columns):
            self.fields["column_{}".format(col_pos)] = forms.ChoiceField(
                choices=settings.ACOMEA_MAILUP_IMPORT_AVAILABLE_FIELDS,
                required=True,
            )

    def column_fields(self):
        for name in self.fields:
            if name.startswith('column_'):
                yield(self[name])

    def clean(self):
        cleaned_data = super(WizardFieldsConfigurationForm, self).clean()
        fields = [value for key, value in cleaned_data.iteritems() if key.startswith('column_') and value != 'ignore']
        for field in fields:
            if fields.count(field) > 1:
                raise forms.ValidationError(_("Field {} selected multiple times".format(field)))
        if 'email' not in fields:
            raise forms.ValidationError(_("Email field undefined"))
        return cleaned_data

    def save(self, commit=True):
        instance = super(WizardFieldsConfigurationForm, self).save(commit=False)
        column_fields = sorted([field for field in self.cleaned_data if field.startswith('column_')])
        instance.fields_configuration = simplejson.dumps(
            [self.cleaned_data[column_field] for column_field in column_fields]
        )
        instance.mailup_list = self.cleaned_data['list']
        instance.status = 'running'
        if commit:
            instance.save()
            msg = _('The task has been queued: you can monitor progress <a href="%s">here<a>')
            messages.info(self.request, mark_safe(msg % reverse('admin:async_task_changelist')))
            ImportAndSyncContactsTask.delay_or_fail(app_label='mailup', instance=instance)
        return instance
