# coding: utf-8
"""
.. _sync-mailup-fields:

Sync Mailup Fields
==================

Command per
usage::

  dj acomea_mailup_contacts_fields_sync_from_csv -f /path/of/file.csv

"""


import csv
import json
import logging
import os
import shutil
import sys
import traceback
from optparse import make_option

import datetime

import psutil
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned
from django.core.mail import send_mail
from django.core.management.base import BaseCommand, CommandError

from mailup.clients import MailUpClientSingleton
from mailup.exceptions import InvalidRecipientConfigurationException
from mailup.providers import MailUpComponentProvider

from acomea.settings.models import MailupFieldMap

# LOGGER
logger = logging.getLogger('sync_contact_fields')


class FieldMappingException(Exception):
    """
    Exception raised when bad mapping of fields is captured
    """
    pass


class UnparseableRowException(Exception):
    """
    Exception raised when row in file is empty or not complete
    """
    pass


# UTILS FUNCTION
def diff_list(list_to_start, list_to_subtract):
    return list(set(list_to_start) - set(list_to_subtract))


class MailUpSynchronizer(object):
    def __init__(self, actual_mailup_list_id=None, **options):
        log_level = getattr(logging,  options.get('debug_level', 'info').upper())
        logger.setLevel(log_level)

        self.start_command_time = datetime.datetime.now()

        # acomea.settings.MailupFieldMap instance
        self.mailup_field_map = MailupFieldMap.objects.all()[0]
        self.filename = settings.MAILUP_FIELDS_SYNC_FILE_NAME

        try:
            # try to open file
            with open(self.filename, 'rb') as csv_file:
                pass
            csv_file.close()
        except Exception as e:
            error_message = 'File {file_name} not found. Error detail {error_detail}'.format(
                    file_name=settings.MAILUP_FIELDS_SYNC_FILE_NAME,
                    error_detail=str(e),
                ),
            logger.critical(error_message)
            send_mail(
                'Acomea MailUp Fields Synchronization {} ERROR'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                error_message,
                'mailupsync@acomea.it',
                settings.MAILUP_FIELDS_SYNC_RECIPIENT_REPORT_MAIL_LIST,
                fail_silently=False,
            )
            sys.exit("Error message")

        if not actual_mailup_list_id:
            raise Exception
        self.actual_mailup_list_id = actual_mailup_list_id
        self.import_id = None
        self.recipient_processed = 0
        self.recipient_updated = 0
        self.recipients_created = 0
        self.recipient_subscribed = 0
        self.recipients_to_subscribe = []
        self.groups_created = 0
        self.row_not_processed_list = []
        self.row_without_email = []
        self.current_row = 1
        self.row_separator = options.get('separator', ';')
        self.lead_updated = 0
        self.header_list = list()

        # mailup client
        self.mailup_client = MailUpClientSingleton(
            client_id=settings.ACOMEA_MAILUP_CLIENT_ID,
            client_secret=settings.ACOMEA_MAILUP_SECRET,
            username=settings.ACOMEA_MAILUP_USERNAME,
            password=settings.ACOMEA_MAILUP_PASSWORD,
        )
        self.mailup_client.configuration_dict['MAILUP_DEFAULT_PAGE_SIZE'] = 3000
        self.mailup_client.configuration_dict['MAILUP_CLIENT_TIMEOUT_403'] = 5

        # Enabling mailup singleton logger to INFO level
        mailup_logger = self.mailup_client.logger
        mailup_logger.disabled = False
        mailup_logger.setLevel(log_level)

        mailup_logger.addHandler(logger.handlers[0])

        # mailup provider
        self.provider = MailUpComponentProvider(client=self.mailup_client, logger=mailup_logger)

        self.mailup_list = self.provider.get_list(self.actual_mailup_list_id)
        # a contact is taken for retrieve fields structure
        logger.info('Retrieve fields structure...')
        sample_contact = self.mailup_list.get_recipients()[0]
        self.fields_structure = sample_contact.fields
        for field in self.fields_structure:
            field['Value'] = ''
        logger.info('Fields structure retrieved')

        # GROUP - RECIPIENT init local
        self.local_groups_name_list = list()
        self.local_recipient_list = list()
        logger.info('local_groups_name_list and local_recipient_list initialization...')
        self.init_local_lists()
        logger.info('local_groups_name_list and local_recipient_list initialized')

        # GROUP - init from mailup
        self.mailup_groups_dict = dict()
        logger.info('mailup_groups_dict initialization...')
        self.init_mailup_groups_dict()
        logger.info('mailup_groups_dict initialized')

        # GROUP - creation on mailup
        logger.info('new group creation on mailup...')
        self.create_new_groups_on_mailup()
        logger.info('new group created on mailup')

        # GROUP - RECIPIENT: participation
        self.local_group_recipient_participation_dict = dict()
        logger.info('local_group_recipient_participation_dict initialization...')
        self.init_local_group_recipient_participation_dict()
        logger.info('local_group_recipient_participation_dict initialized')

        # RECIPIENT - init from mailup
        self.mailup_recipients_dict = dict()
        logger.info('mailup_recipients_dict initialization...')
        self.init_mailup_recipients_dict()
        logger.info('mailup_recipients_dict initialized')

        # RECIPIENT - creation on mailup
        self.create_new_recipients_on_mailup()

        super(MailUpSynchronizer, self).__init__()

    # GROUP - RECIPIENT init local
    def init_local_lists(self):
        self.local_groups_name_list = list()
        self.local_recipient_list = list()

        if self.filename is not None:
            self.current_row = 1
            with open(self.filename, 'rb') as csv_file:

                # Read file line by line
                file_reader = csv.reader(csv_file, delimiter=self.row_separator, quotechar='"')
                for index, line in enumerate(file_reader):
                    # build row_dict from file row
                    if self.current_row == 1:
                        row = self.row_separator.join(line).decode('utf-8-sig')
                        row_list = row.split(self.row_separator)
                        self.header_list = row_list
                        self.current_row += 1
                        # header row is skipped
                        continue
                    row = self.row_separator.join(line)
                    row_list = row.split(self.row_separator)
                    try:
                        row_dict = self.build_row_dict(row_list)
                    except UnparseableRowException:
                        continue

                    # recipient append
                    if not row_dict['Email']:
                        logger.warning('Row without email found in {}'.format(self.filename))
                        self.row_without_email.append(index + 1)

                    email = row_dict['Email']
                    if email:
                        self.local_recipient_list.append(email.lower().strip())

                    # group append
                    for group_name in row_dict['Gruppi'].split(','):
                        if group_name:
                            self.local_groups_name_list.append(group_name.strip())

                    self.current_row += 1

                # group distinct
                self.local_groups_name_list = list(set(self.local_groups_name_list))

        else:
            logger.critical('File not found')

    # GROUP - RECIPIENT: participation
    def init_local_group_recipient_participation_dict(self):
        for group_name in self.local_groups_name_list:
            group = self.provider.filter_groups(self.mailup_list.id, filters={'Name': group_name})[0]
            recipients_list = group.get_subscribers()
            recipients_email_list = [recipient.email for recipient in recipients_list]
            self.local_group_recipient_participation_dict[group_name] = recipients_email_list

    # GROUP - init from mailup
    def init_mailup_groups_dict(self):
        mailup_groups_list = self.provider.all_groups(self.mailup_list.id)
        mailup_groups_dict = {group.name: group for group in mailup_groups_list}
        self.mailup_groups_dict = mailup_groups_dict

    # GROUP - creation on mailup
    def create_new_groups_on_mailup(self):
        groups_name_list = diff_list(self.local_groups_name_list, self.mailup_groups_dict.keys())
        self.groups_created = len(groups_name_list)
        for group_name in groups_name_list:
            group = self.provider.create_group(
                {
                    'idList': self.mailup_list.id,
                    'Name': group_name,
                    'Notes': 'Automatically created from "acomea_mailup_contacts_fields_sync_from_csv" scedulated command',
                }
            )
            self.mailup_groups_dict[group.name] = group

    def save_recipients_json_backup(self, mailup_recipients_list):
        # recipient_json_file_name = settings.MAILUP_FIELDS_SYNC_RECIPIENT_FILE_BACKUP_PATH + \
        #                            'recipients_{}.json'.format(
        #                                datetime.datetime.now().strftime('%Y%m%d_%H%M')
        #                            )
        if self.mailup_list and self.mailup_list.id:
            list_id = self.mailup_list.id
        else:
            list_id = 'NA'
        recipient_json_file_name = settings.MAILUP_FIELDS_SYNC_RECIPIENT_FILE_BACKUP_PATH + 'list-{}-recipients.json'.format(list_id)
        # recipient_json_file_name = settings.MAILUP_FIELDS_SYNC_RECIPIENT_FILE_BACKUP_PATH + 'recipients.json'
        logger.info('recipient backup creation at {}'.format(recipient_json_file_name))

        # Build list to serialize
        list_to_serialize = []
        for rec in mailup_recipients_list:
            list_to_serialize.append(rec.data_dict)

        # json_string is saved on file
        recipient_json_file = open(recipient_json_file_name, "w")
        json.dump(list_to_serialize, recipient_json_file)
        recipient_json_file.close()

    def backup_contact_table(self):
        # TODO: fare backup dei contatti
        pass

    # RECIPIENT - init from mailup
    def init_mailup_recipients_dict(self):
        mailup_recipients_list = self.provider.all_recipients(self.mailup_list.id)

        # backup
        if settings.MAILUP_FIELDS_SYNC_RECIPIENT_FILE_BACKUP_ENABLED:
            self.save_recipients_json_backup(mailup_recipients_list)
            self.backup_contact_table()

        mailup_recipients_dict = {recipient.email.lower(): recipient for recipient in mailup_recipients_list}
        self.mailup_recipients_dict = mailup_recipients_dict

    # RECIPIENT - creation on mailup
    def create_new_recipients_on_mailup(self):
        emails_list = diff_list(self.local_recipient_list, self.mailup_recipients_dict.keys())
        self.recipients_created = len(emails_list)
        for email in emails_list:
            try:
                recipient = self.provider.create_or_update_recipient(
                    {
                        'idList': self.mailup_list.id,
                        'Name': '',
                        'Email': email,
                    }
                )
                # new recipient is provided of an empty fields structure
                recipient.data_dict['Fields'] = [dict(t) for t in set([tuple(d.items()) for d in self.fields_structure])]
                self.mailup_recipients_dict[recipient.email] = recipient
            except InvalidRecipientConfigurationException as e:
                error_string = '{exception_name}: {error_text}'.format(
                    exception_name=e.__class__.__name__,
                    error_text=e.error_text,
                )
                logger.error('Mailup response after recipient "{email}" creation without data_dict: {error_string}'.format(
                    email=email,
                    error_string=error_string,
                ))
            except Exception as e:
                error_string = '{exception_name}: {error_text}'.format(
                    exception_name=e.__class__.__name__,
                    error_text=e.message,
                )
                logger.error('Generic Error during create recipient "{email}": {error_string}'.format(
                    email=email,
                    error_string=error_string,
                ))

    # SYNCHRONIZATION FILE->MAILUP
    def do_sync(self):
        if self.filename is not None:
            logger.info('\t###MAILUP SYNCHRONIZATION START###')
            self.current_row = 1
            with open(self.filename, 'rb') as csv_file:

                # Read file line by line
                file_reader = csv.reader(csv_file, delimiter=self.row_separator, quotechar='"')
                num_lines = sum(1 for line in open(self.filename, 'rb'))
                for line in file_reader:
                    # build row_dict from file row

                    if self.current_row == 1:
                        row = self.row_separator.join(line).decode('utf-8-sig')
                        row_list = row.split(self.row_separator)
                        self.header_list = row_list
                        self.current_row += 1
                        # header row is skipped
                        continue

                    row = self.row_separator.join(line)
                    row_list = row.split(self.row_separator)
                    try:
                        row_dict = self.build_row_dict(row_list)
                    except UnparseableRowException:
                        self.row_not_processed_list.append(self.current_row)
                        continue

                    self.update_first_last_if_lead(
                        email=unicode(row_list[0], 'utf-8'),
                        first_name=unicode(row_list[1], 'utf-8'),
                        last_name=unicode(row_list[2], 'utf-8'),
                    )

                    # process row_dict
                    try:
                        result = None
                        recipient, result = self.process_row(
                            row_dict,
                            email=unicode(row_list[0].lower(), 'utf-8'),
                            first_name=unicode(row_list[1], 'utf-8'),
                            last_name=unicode(row_list[2], 'utf-8'),
                        )
                    except Exception as e:
                        logger.error('Error during processing row {row}: {message}'.format(
                            row=self.current_row,
                            message=e,
                        ))
                        if not result:
                            result = 'not_updated_recipient'

                    # counts
                    if result == 'recipient_processed':
                        self.recipient_processed += 1
                    if result == 'recipient_updated':
                        self.recipient_processed += 1
                        self.recipient_updated += 1
                    if result == 'not_updated_recipient':
                        self.row_not_processed_list.append(self.current_row)

                    # current row increment
                    self.current_row += 1
                    logger.info('Lista: {}\tProcessed:{}/{}'.format(
                        self.actual_mailup_list_id,
                        self.recipient_processed,
                        num_lines - 1
                    ))

            # recipients subscription
            if (
                (self.mailup_list.id == settings.ACOMEA_GIMME5_MAILUP_LIST_ID and settings.MAILUP_GIMME5_FIELDS_SYNC_SUBSCRIBE_RECIPIENT)
                or (self.mailup_list.id != settings.ACOMEA_GIMME5_MAILUP_LIST_ID  and settings.MAILUP_FIELDS_SYNC_SUBSCRIBE_RECIPIENT)
            ) and len(self.recipients_to_subscribe):
                self.import_id = self.mailup_list.subscribe_recipients_list_forced(self.recipients_to_subscribe)
                self.recipient_subscribed = len(self.recipients_to_subscribe)

            # end-import result
            import_result = """\n
\tTOTAL RESULT FOR LIST ID {mailup_list_id}:
            - processed row:....................{total_row}
            - not processed row line:...........{row_not_processed_list}
            - row without email:................{row_without_email}
            - recipients updated:...............{updated_recipient}
            - recipients not updated:...........{updated_not_recipient}
            - recipients created:...............{created_recipient}
            - lead updated (first and last):....{lead_updated}
            - groups created:...................{created_group}
            - filename:.........................{filename}
            - subscribed recipient:.............{subscribed_recipient}   (import_id: {import_id})

\tCONFIGURATION:
            - mailup recipient backup enabled:..{backup_enabled}
            - subscription enabled:.............{subscribe_recipient}

        execution time: {execution_time}
        """.format(
                mailup_list_id=self.actual_mailup_list_id,
                filename=self.filename,
                total_row=num_lines - 1,
                created_recipient=self.recipients_created,
                lead_updated=self.lead_updated,
                updated_recipient=self.recipient_updated,
                created_group=self.groups_created,
                row_not_processed_list=self.row_not_processed_list,
                row_without_email=self.row_without_email,
                updated_not_recipient=len(self.row_not_processed_list),
                execution_time=str(datetime.datetime.now() - self.start_command_time),
                backup_enabled=settings.MAILUP_FIELDS_SYNC_RECIPIENT_FILE_BACKUP_ENABLED,
                subscribe_recipient=settings.MAILUP_FIELDS_SYNC_SUBSCRIBE_RECIPIENT,
                subscribed_recipient=self.recipient_subscribed,
                import_id=self.import_id,
            )
            logger.info(import_result)
            send_mail(
                'Acomea MailUp Fields Synchronization report {}'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                '{}'.format(import_result),
                'mailupsync@acomea.it',
                settings.MAILUP_FIELDS_SYNC_RECIPIENT_REPORT_MAIL_LIST,
                fail_silently=False,
            )
            logger.info('\t###MAILUP SYNCHRONIZATION END###\n\n')
            return import_result
        else:
            logger.critical('File not found')

    def process_row(self, row_dict, email, first_name, last_name):
        recipient = self.mailup_recipients_dict[email]
        recipient.name = u'{} {}'.format(first_name, last_name)
        to_save, recipient.fields = self.update_fields_groups(row_dict, recipient)

        if not recipient.status == 'subscribed':
            logger.info('recipient_id {} with email {} will be subscribed in list'.format(recipient.id, recipient.email))
            self.recipients_to_subscribe.append(recipient)

        if to_save:
            recipient.save()
            logger.info('Recipient fields updated')
            return recipient, 'recipient_updated'
        else:
            logger.info('Recipient fields already updated, save bypassed')
            return recipient, 'recipient_processed'

    def update_fields_groups(self, row_dict, recipient):
        """
        Update recipient from csv row, recipient must be saved after this execution
        :param row_dict:
        :param recipient:
        :return:
        """
        to_save = False
        for column_header, column_value in row_dict.items():
            column_value = unicode(column_value, 'utf-8')
            if column_header == 'Email':
                continue
            if column_header == 'Nome':
                if not recipient.get_field('Nome') == column_value:
                    logger.info(
                        u'recipient {recipient_id} will be saved because field "{field_name}" has different values: '
                        u'old_value: "{old_value}" new_value: "{new_value}"'.format(
                            recipient_id=recipient.id,
                            field_name='Nome',
                            old_value=recipient.get_field('Nome'),
                            new_value=column_value,
                        )
                    )
                    recipient.set_field('Nome', column_value)
                    to_save = True
                continue
            if column_header == 'Cognome':
                if not recipient.get_field('Cognome') == column_value:
                    logger.info(
                        u'recipient {recipient_id} will be saved because field "{field_name}" has different values: '
                        u'old_value: "{old_value}" new_value: "{new_value}"'.format(
                            recipient_id=recipient.id,
                            field_name='Cognome',
                            old_value=recipient.get_field('Cognome'),
                            new_value=column_value,
                        )
                    )
                    recipient.set_field('Cognome', column_value)
                    to_save = True
                continue
            if column_header == 'Gruppi':
                for group_name in column_value.split(','):
                    group_name = group_name.strip()
                    if group_name:
                        # check if recipient is in group
                        if recipient.email not in self.local_group_recipient_participation_dict[group_name]:
                            # recipient is inserted into group
                            recipient.add_to_group(self.mailup_groups_dict[group_name].id)
                continue

            try:
                mailup_field_name = self.get_mailup_field_name_from_mapping(column_header)
                if not recipient.get_field(mailup_field_name) is None:
                    # if recipient has ho mailup_field_name, then recipient.get_field(mailup_field_name) is None
                    if not recipient.get_field(mailup_field_name) == column_value:
                        logger.info(
                            u'recipient {recipient_id} will be saved because field "{field_name}" has different values:'
                            u' old_value: "{old_value}" new_value: "{new_value}"'.format(
                                recipient_id=recipient.id,
                                field_name=mailup_field_name,
                                old_value=recipient.get_field(mailup_field_name),
                                new_value=column_value,
                            )
                        )
                        recipient.set_field(mailup_field_name, column_value)
                        to_save = True
            except FieldMappingException:
                pass

        return to_save, recipient.fields

    # UTILS
    def build_row_dict(self, row_list):
        row_dict = dict()
        try:
            for index, value in enumerate(self.header_list):
                row_dict[value] = row_list[index]
        except IndexError:
            error_msg = 'Unparseable row: %s' % row_list
            logger.error(error_msg)
            raise UnparseableRowException(error_msg)
        return row_dict

    def update_first_last_if_lead(self, email, first_name, last_name):
        from django.contrib.auth.models import User

        try:
            user = User.objects.get(email=email)
        except User.DoesNotExist:
            logger.warning('Email {} has no user in acomea DB'.format(email))
            return
        except MultipleObjectsReturned:
            logger.warning('Email {} has multiple user in acomea DB'.format(email))
            return

        if hasattr(user, 'profilo') and user.profilo.is_only_lead and hasattr(user.profilo, 'stato') and user.profilo.stato == '1':
            # aggiorno solo i profili is_only_lead in stato nessuna richiesta

            self.lead_updated += 1

            # ONLY LEAD USER IS UPDATED
            user.first_name = first_name
            user.last_name = last_name
            user.save()
            if hasattr(user, 'contact'):
                user.contact.first_name = first_name
                user.contact.last_name = last_name
                user.contact.save()
            logger.info('lead "{}" first_name and last_name updated'.format(email))

    def get_dict_item_list_from_value(self, dictionary, value):
        """
        Utility, example:
        :param dictionary:{'foo': 'f', 'bar', b}
        :param value: 'foo'
        :return: [('foo', u'f')]
        """
        item_list = list()
        for item in dictionary.iteritems():
            if item[1] == value and 'export_file' in item[0]:
                item_list.append(item)
        return item_list

    def get_mailup_field_name_from_mapping(self, csv_header):
        """
        Calculate mailup_field_name from csv_header using DB configuration (self.mailup_field_map)
        :param csv_header: Header of column
        :return: mailup_field_name
        """
        field_item_list = self.get_dict_item_list_from_value(self.mailup_field_map.__dict__, csv_header)
        if not len(field_item_list) == 1:
            raise FieldMappingException
        file_map_field = field_item_list[0]
        item_position = file_map_field[0].split('_')[1]
        map_field_name = 'field_{}_mailup'.format(item_position)
        mailup_field_name = self.mailup_field_map.__dict__[map_field_name]
        return mailup_field_name


class Command(BaseCommand):

    option_list = BaseCommand.option_list + (
        # -c enables logging using store_true
        make_option(
            '-c', '--console', action='store_true', default=False, help='Debug - write logging to console'
        ),
        # file selector (required)
        make_option('-f', '--file', dest='filename', help='File to process', default=None),
        # debug level in console
        make_option('-l', '--debug-level', default='debug',
                    help='Set debug level (debug, info, warnings) for console', ),
        # CSV separator
        make_option(
            '-s', '--separator', dest='separator', type='str',
            help='Set row separator, default is ";"', default=';'
        ),
    )

    output_transaction = True
    actual_mailup_list_id = None

    def handle(self, *args, **options):
        # itself pid and parent pid
        pid_to_exclude = [os.getpid(), os.getppid()]
        procs = [proc for proc in psutil.process_iter() if proc.name() == 'django' and proc.pid not in pid_to_exclude]
        for proc in procs:
            if 'acomea_mailup_contacts_fields_sync_from_csv' in proc.cmdline():
                logger.error('Script precedente ancora in esecuzione.')
                send_mail(
                    'acomea mailup contacts fields sync from csv',
                    'Script precedente ancora in esecuzione',
                    'mailupsync@acomea.it', settings.MAILUP_FIELDS_SYNC_RECIPIENT_REPORT_MAIL_LIST,
                    fail_silently=False
                )
                raise CommandError('Script precedente ancora in esecuzione')
        filename = '{}_{}'.format(
            datetime.datetime.now().strftime('%Y%m%d_%H%M'),
            settings.MAILUP_FIELDS_SYNC_FILE_NAME.split('/')[-1],
        )
        try:
            for mailup_list_id in [settings.ACOMEA_MAILUP_LIST_ID, settings.ACOMEA_GIMME5_MAILUP_LIST_ID]:
                self.actual_mailup_list_id = mailup_list_id
                synchronizer = MailUpSynchronizer(actual_mailup_list_id=mailup_list_id, **options)
                synchronizer.do_sync()

            # mv processed file
            shutil.move(settings.MAILUP_FIELDS_SYNC_FILE_NAME, settings.MAILUP_FIELDS_SYNC_PROCESSED_FILES + filename)
        except Exception as e:
            # saving backup file
            shutil.copy(settings.MAILUP_FIELDS_SYNC_FILE_NAME, settings.MAILUP_FIELDS_SYNC_PROCESSED_FILES + filename + '.KO')

            # send error email
            error_message = traceback.format_exc()

            logger.critical(error_message)
            send_mail(
                'Acomea MailUp Field Synchronization {} ERROR'.format(datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")),
                error_message,
                'mailupsync@acomea.it',
                settings.MAILUP_FIELDS_SYNC_RECIPIENT_REPORT_MAIL_LIST,
                fail_silently=False,
            )
            sys.exit("Error message")
