# -*- coding: utf-8 -*-
"""Command for contacts sync in mailup console"""

import datetime
import time
import logging
from optparse import make_option

from django.conf import settings
from django.core.management.base import BaseCommand

from acomea.mailup.utils.client import MailUpClient
from acomea.contact.models import Contact, Group, Email

log = logging.getLogger("mailup")


class Command(BaseCommand):
    """..."""
    help = 'Test mailup'
    option_list = BaseCommand.option_list + (
        make_option('-t', '--today', action="store_true", dest='today', help='Created today', default=False),
        make_option('-r', '--reverse', action="store_true", dest='reverse', help='Reverse sync', default=False),
        make_option('-g', '--group', dest='group', help='Group', default=None),
    )
    mailup_client = MailUpClient()
    list_id = settings.ACOMEA_MAILUP_LIST_ID
    mailup_group_recipients = {}
    emails_done = []
    read_groups_json = None

    def handle(self, *scripts, **opts):
        list_id = self.list_id
        reverse = opts.get('reverse')
        if list_id:
            log.debug("### INIZIO SINCRONIZZAZIONE CONTATTI IN LISTA %s ###" % list_id)
            self.mailup_client.retrieve_access_token()
            self.get_mailup_group_recipients()
            self.manage_contacts_from_db(opts)
            if reverse and datetime.date.today() == 5:
                self.manage_contacts_from_mailup()
        else:
            log.debug("Specificare l'id della lista mailup nella quale effettuare la sincronizzazione dei contatti")

    def get_mailup_group_recipients(self):
        self.read_groups_json = self.mailup_client.read_groups(self.list_id)
        mailup_groups = [item["idGroup"] for item in self.read_groups_json["Items"]]
        Group.objects.exclude(
            mailup_group_id__in=mailup_groups
        ).filter(mailup_group_id__isnull=False).update(mailup_group_id=None)
        for mailup_group in mailup_groups:
            self.mailup_group_recipients[mailup_group] = []
            read_group_members_json = self.mailup_client.read_group_members(mailup_group)
            if read_group_members_json:
                for recipient in read_group_members_json["Items"]:
                    if recipient["idRecipient"] not in self.mailup_group_recipients[mailup_group]:
                        self.mailup_group_recipients[mailup_group].append(recipient["idRecipient"])
            log.debug("MailUP Group %s - contatti trovati %s" % (
                mailup_group, len(self.mailup_group_recipients[mailup_group])
            ))

    def manage_contacts_from_db(self, opts):
        list_id = self.list_id
        today = opts.get('today')
        group_id = opts.get('group')
        contact_cont = 1
        qs = Contact.objects.all()

        # Filtro per data
        if today:
            date_filter = datetime.date.today()-datetime.timedelta(days=2)
            qs = qs.filter(date_last_modify_for_mailup__gte=date_filter)
            log.debug("Data filtro ultima modifica contatti: %s" % date_filter)

        # Filtro per gruppo
        if group_id:
            group = Group.objects.get(id=group_id)
            qs = qs.filter(groups=group)

        total = qs.count()
        log.debug("Contatti da aggiornare da database: %s" % total)
        # Ciclo i contatti del database
        for contact in qs:
            contact_email = contact.get_default_email()
            log.debug("###################################################################")
            log.debug("Inizio sincronizzazione contatto: %s (%s/%s)" % (contact_email, contact_cont, total))
            if contact_email and contact_email.strip():
                try:
                    # Importo il contatto singolo
                    if not contact.mailup_recipient_id:
                        mailup_recipient_id = self.mailup_client.import_recipient(list_id, contact)
                        contact.mailup_recipient_id = mailup_recipient_id
                        contact.save()
                    else:
                        mailup_recipient_id = contact.mailup_recipient_id
                    log.debug("Mailup recipient id: %s" % mailup_recipient_id)
                    self.emails_done.append(contact_email.lower().strip())
                    # Iscrizione/Disiscrizione
                    status_dict = {True: "Subscribed", False: "Unsubscribed"}
                    is_subscribed, msg = contact.can_create_message
                    db_status = status_dict[is_subscribed]
                    is_subscribed_json = self.mailup_client.check_subscriber_exists(list_id, "Subscribed", contact_email)
                    if not is_subscribed_json:
                        mailup_status = "Subscribed" if db_status == "Unsubscribed" else "Unsubscribed"
                    else:
                        mailup_status = "Subscribed" if is_subscribed_json["TotalElementsCount"] else "Unsubscribed"
                    if db_status != mailup_status:
                        if db_status == "Subscribed":
                            self.mailup_client.import_recipients(list_id, [contact], "asOptin")
                            log.debug("Contatto iscritto")
                        elif db_status == "Unsubscribed":
                            self.mailup_client.unsubscribe_recipient(list_id, mailup_recipient_id)
                            log.debug("Contatto disiscritto")
                    else:
                        log.debug("Stato contatto gia' allineato")
                    if db_status == "Subscribed":
                        contact_groups = self.get_contact_groups(contact, mailup_recipient_id)
                        if contact_groups['db_groups'] != contact_groups['mailup_groups']:
                            self.manage_contact_groups_from_db(contact, mailup_recipient_id)
                            self.manage_contact_groups_from_mailup(
                                contact, mailup_recipient_id, contact_groups['mailup_groups']
                            )
                        else:
                            log.debug("Elenco gruppi contatto gia allineato")
                    else:
                        log.debug("Non aggiorno gruppi dato che disiscritto")
                except Exception, e:
                    log.debug("Eccezione in sincronizzazione da database per %s: %s" % (contact_email, e))
            contact_cont += 1

    def manage_contacts_from_mailup(self):
        for status in ["Subscribed", "Unsubscribed"]:
            items = self.mailup_client.read_list_recipients(self.list_id, status)['Items']
            recipients = [recipient for recipient in items if recipient['Email'].lower().strip() not in self.emails_done]
            log.debug("###################################################################")
            log.debug("Contatti da aggiornare in stato %s da mailup: %s" % (status, len(recipients)))
            contact_cont = 1
            for recipient in recipients:
                try:
                    log.debug("###################################################################")
                    email = recipient['Email']
                    mailup_recipient_id = recipient['idRecipient']
                    first_name = [field['Value'] for field in recipient['Fields'] if field['Description'] == 'nome'][0]
                    last_name = [field['Value'] for field in recipient['Fields'] if field['Description'] == 'cognome'][0]
                    log.debug("Inizio sincronizzazione contatto: %s (%s/%s)" % (email, contact_cont, len(recipients)))
                    contact_exists = Contact.objects.filter(emails__email__exact=email.lower()).exists()
                    contact = None
                    if contact_exists:
                        contact = Contact.objects.get(emails__email__exact=email.lower())
                        if contact.mailup_recipient_id != mailup_recipient_id:
                            contact.mailup_recipient_id = mailup_recipient_id
                            contact.save()
                        log.debug("Contatto gia' esistente")
                    elif not contact_exists and status == "Subscribed":
                        contact = Contact.objects.create(
                            first_name=first_name,
                            last_name=last_name,
                            mailup_recipient_id=mailup_recipient_id,
                            bonifica=True,
                            auth_trattamento_dati=True,
                            auth_materiale_commerciale=True,
                        )
                        Email.objects.create(contact=contact, email=email, default=True)
                        log.debug("Contatto creato")
                    elif not contact_exists and status == "Unsubscribed":
                        log.debug("Contatto non presente su database")
                    if contact:
                        # Controllo stato newsletter contatto
                        status_dict = {True: "Subscribed", False: "Unsubscribed"}
                        is_subscribed, msg = contact.can_create_message
                        db_status = status_dict[is_subscribed]
                        if db_status != status:
                            if status == "Subscribed":
                                contact.auth_materiale_commerciale = True
                                log.debug("Contatto abilitato")
                            elif status == "Unsubscribed":
                                contact.auth_materiale_commerciale = False
                                log.debug("Contatto disabiliato")
                            contact.save()
                        else:
                            log.debug("Stato contatto gia' allineato")
                        # Gestisco i gruppi se necessario
                        contact_groups = self.get_contact_groups(contact, mailup_recipient_id)
                        if contact_groups['db_groups'] != contact_groups['mailup_groups']:
                            self.manage_contact_groups_from_db(contact, mailup_recipient_id)
                            self.manage_contact_groups_from_mailup(
                                contact, mailup_recipient_id, contact_groups['mailup_groups']
                            )
                        else:
                            log.debug("Elenco gruppi contatto gia allineato")
                    contact_cont += 1
                except Exception, e:
                    log.debug("Eccezione in sincronizzazione da mailup per %s: %s" % (email, e))
                    contact_cont += 1

    def get_contact_groups(self, contact, mailup_recipient_id):
        return {
            'db_groups': sorted(
                contact.groups.exclude(mailup_group_id__isnull=True).values_list(
                    'mailup_group_id', flat=True
                )
            ),
            'mailup_groups': sorted([
                group for group, group_recipients in self.mailup_group_recipients.iteritems()
                if mailup_recipient_id in group_recipients
            ])
        }

    def manage_contact_groups_from_db(self, contact, mailup_recipient_id):
        for group in contact.groups.all():
            if not group.mailup_group_id:
                create_group_json = self.mailup_client.create_group(self.list_id, group.name)
                group.mailup_group_id = create_group_json["idGroup"]
                group.save()
            else:
                self.mailup_client.update_group(self.list_id, group.mailup_group_id, group.name)
            if group.mailup_group_id in self.mailup_group_recipients \
                    and mailup_recipient_id in self.mailup_group_recipients[group.mailup_group_id]:
                continue
            self.mailup_client.assign_recipient_to_group(group.mailup_group_id, mailup_recipient_id)
            if group.mailup_group_id not in self.mailup_group_recipients:
                self.mailup_group_recipients[group.mailup_group_id] = []
            if mailup_recipient_id not in self.mailup_group_recipients[group.mailup_group_id]:
                self.mailup_group_recipients[group.mailup_group_id].append(mailup_recipient_id)
            log.debug("Contatto assegnato al gruppo in %s in mailup" % group.mailup_group_id)

    def manage_contact_groups_from_mailup(self, contact, mailup_recipient_id, mailup_groups):
        for mailup_group_id in mailup_groups:
            try:
                group = Group.objects.get(mailup_group_id=mailup_group_id)
            except Group.DoesNotExist:
                mailup_group_id, mailup_group_name = [
                    (group["idGroup"], group["Name"])
                    for group in self.read_groups_json["Items"] if int(group["idGroup"]) == mailup_group_id
                ][0]
                group = Group(name=mailup_group_name, mailup_group_id=mailup_group_id)
                group.save()
            if group not in contact.groups.all():
                contact.groups.add(group)  # signal "contact_group_signal" change "first_group" field, to use contact instace remember to reload
                log.debug(
                    "Contatto assegnato al gruppo %s su database (mailup_id: %s)" % (group.id, group.mailup_group_id)
                )
