# -*- coding: utf-8 -*-
"""
helper module for acomea.mailup
"""

import base64
import json
import requests
import time
import logging

from django.conf import settings


log = logging.getLogger("mailup")

class MailUpClient():

    def __init__(self):
        self.logonEndpoint = "https://services.mailup.com/Authorization/OAuth/LogOn"
        self.authorizationEndpoint = "https://services.mailup.com/Authorization/OAuth/Authorization"
        self.tokenEndpoint = "https://services.mailup.com/Authorization/OAuth/Token"
        self.consoleEndpoint = "https://services.mailup.com/API/v1.1/Rest/ConsoleService.svc"
        self.mailStatisticsEndpoint = "https://services.mailup.com/API/v1.1/Rest/MailStatisticsService.svc"

        self.clientId = settings.ACOMEA_MAILUP_CLIENT_ID
        self.clientSecret = settings.ACOMEA_MAILUP_SECRET
        self.accessToken = None
        self.refreshToken = None

    def retrieve_access_token(self):
        url = self.tokenEndpoint
        params = {
            "grant_type": "password",
            "client_id": self.clientId,
            "client_secret": self.clientSecret,
            "username": settings.ACOMEA_MAILUP_USERNAME,
            "password": settings.ACOMEA_MAILUP_PASSWORD,
        }
        headers = {
            'content-type': 'application/x-www-form-urlencoded',
            'Authorization': "Basic %s" % base64.b64encode(self.clientId+":"+self.clientSecret)
        }
        rest_request_json = self.call_method("POST", url, params=params, headers=headers)
        self.accessToken = rest_request_json["access_token"]
        self.refreshToken = rest_request_json["refresh_token"]

    def refresh_token(self):
        url = self.tokenEndpoint
        params = {
            "grant_type": "refresh_token",
            "client_id": self.clientId,
            "client_secret": self.clientSecret,
            "refresh_token": self.refreshToken,
        }
        headers = {
            'content-type': 'application/x-www-form-urlencoded',
        }
        rest_request_json = self.call_method("POST", url, params=params, headers=headers)
        self.accessToken = rest_request_json["access_token"]
        self.refreshToken = rest_request_json["refresh_token"]

    def read_authentication_info(self):
        url = self.consoleEndpoint + "/Console/Authentication/Info"
        headers = {"Authorization": "Bearer %s" % self.accessToken}
        return self.call_method("GET", url, headers=headers)

    def read_lists(self, page_number=0, page_size=20):
        params = {"PageNumber":page_number, "PageSize": page_size}
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        url = self.consoleEndpoint + "/Console/User/Lists"
        return self.call_method("GET", url, params=params,  headers=headers)

    def create_group(self, list_id, group_name):
        url = self.consoleEndpoint + "/Console/List/%s/Group" % list_id
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        data = {
            "Deletable": True,
            "Name": "%s" % group_name,
            "Notes": "",
            "idList": "%s" % list_id,
        }
        return self.call_method("POST", url, data=json.dumps(data), headers=headers)

    def read_groups(self, list_id):
        url = self.consoleEndpoint + "/Console/List/%s/Groups" % list_id
        params = {"PageNumber":0, "PageSize": 999999}
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_group_members(self, group_id):
        url = self.consoleEndpoint + "/Console/Group/%s/Recipients" % group_id
        params = {"PageNumber": 0, "PageSize": 999999}
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params,  headers=headers)

    def read_list_recipients(self, list_id, status):
        url = self.consoleEndpoint + "/Console/List/%s/Recipients/%s" % (list_id, status)
        params = {"PageNumber": 0, "PageSize": 999999}
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params,  headers=headers)

    def update_group(self, list_id, group_id, group_name, group_notes=""):
        url = self.consoleEndpoint + "/Console/List/%s/Group/%s" % (list_id, group_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        data = {
            "Name": "%s" % group_name,
            "Notes": "%s" % group_notes,
            "idGroup": "%s" % group_id,
            "idList": "%s" % list_id
        }
        return self.call_method("PUT", url, data=json.dumps(data), headers=headers)

    def delete_group(self, list_id, group_id):
        url = self.consoleEndpoint + "/Console/List/%s/Group/%s" % (list_id, group_id)
        headers = {
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("DELETE", url, headers=headers, method_type="raw")

    def assign_recipient_to_group(self, group_id, recipient_id):
        url = self.consoleEndpoint + "/Console/Group/%s/Subscribe/%s" % (group_id, recipient_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("POST", url, headers=headers, method_type="raw")

    def remove_recipient_from_group(self, group_id, recipient_id):
        url = self.consoleEndpoint + "/Console/Group/%s/Unsubscribe/%s" % (group_id, recipient_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("DELETE", url, headers=headers, method_type="raw")

    def check_subscriber_exists(self, list_id, status, email):
        url = self.consoleEndpoint + "/Console/List/%s/Recipients/%s" % (list_id, status)
        headers = {
            "Authorization": "Bearer %s" % self.accessToken,
        }
        params = {
            "filterby": "Email.Contains('%s')" % email,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def subscribe_recipient(self, list_id, recipient_id):
        url = self.consoleEndpoint + "/Console/List/%s/Subscribe/%s" % (list_id, recipient_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("POST", url, headers=headers, method_type="raw")

    def unsubscribe_recipient(self, list_id, recipient_id):
        url = self.consoleEndpoint + "/Console/List/%s/Unsubscribe/%s" % (list_id, recipient_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("DELETE", url, headers=headers, method_type="raw")

    def import_recipient(self, list_id, contact):
        url = self.consoleEndpoint + "/Console/List/%s/Recipient" % list_id
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        data = {
            "Email": contact.get_default_email().strip(),
            "Name": "",
            "Fields": [
                {
                    "Description": "FirstName",
                    "Id": 1,
                    "Value": "%s" % contact.first_name,
                },
                {
                    "Description": "LastName",
                    "Id": 2,
                    "Value": "%s" % contact.last_name,
                },
                {
                    "Id": 10,
                    "Value": "%s" % contact.id,
                },
            ],
        }
        return self.call_method("POST", url, data=json.dumps(data), headers=headers)

    def import_recipients_in_group(self, group_id, contacts):
        url = self.consoleEndpoint + "/Console/Group/{}/Recipients".format(group_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        data = []
        for contact in contacts:
            data.append(
                {
                    "Email": contact.get_default_email().strip(),
                    "Name": "",
                    "Fields": [
                        {
                            "Description": "FirstName",
                            "Id": 1,
                            "Value": contact.first_name,
                        },
                        {
                            "Description": "LastName",
                            "Id": 2,
                            "Value": contact.last_name,
                        },
                        {
                            "Id": 10,
                            "Value": contact.id,
                        },
                    ],
                }
            )
        return self.call_method("POST", url, data=json.dumps(data), headers=headers)

    def import_recipients(self, list_id, contacts, import_type):
        url = self.consoleEndpoint + "/Console/List/{}/Recipients".format(list_id)
        params = {"importType": import_type}
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        data = [{
            "Email": contact.get_default_email().strip(),
            "Name": "",
            "Fields": [
                {
                    "Description": "FirstName",
                    "Id": 1,
                    "Value": "%s" % contact.first_name,
                },
                {
                    "Description": "LastName",
                    "Id": 2,
                    "Value": "%s" % contact.last_name,
                },
                {
                    "Id": 10,
                    "Value": "%s" % contact.id,
                },
            ],
        } for contact in contacts]
        return self.call_method("POST", url, params=params, data=json.dumps(data), headers=headers)

    def read_import_status(self, import_id):
        url = self.consoleEndpoint + "/Console/Import/%s" % import_id
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, headers=headers)

    def read_list_messages(self, list_id, page_number=0, page_size=20):
        """
        Used to return all messages of a specified list
        :param list_id: Mailup list id you need to analyze
        :return: {"IsPaginated":false,"Items":[
            {"Subject":"test","idList":2,"idMessage":29}
        ],"PageNumber":0,"PageSize":20,"Skipped":0,"TotalElementsCount":4}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.consoleEndpoint + "/Console/List/{}/Emails".format(list_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_message_recipients(self, message_id, page_number=0, page_size=20):
        """
        Used to list all recipient for a specified message
        :param message_id: Mailup message id you need to analyze
        :return: {"IsPaginated":true,"Items":[
            {"Email":"george@example.com","IdMessage":11826,"IdRecipient":1376600},
        ],"PageNumber":0,"PageSize":20,"Skipped":0,"TotalElementsCount":54}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.mailStatisticsEndpoint + "/Message/{}/List/Recipients".format(message_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_message_openings(self, message_id, page_number=0, page_size=20):
        """
        Used to list all recipient openings for a specified message
        :param message_id: Mailup message id you nedd to analyze
        :return: {"IsPaginated":false,"Items":[
            {"Count":3,"Email":"ken@example.com","IdMessage":11826,"IdRecipient":303812}
        ],"PageNumber":0,"PageSize":20,"Skipped":0,"TotalElementsCount":8}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.mailStatisticsEndpoint + "/Message/{}/List/Views".format(message_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_message_bounces(self, message_id, page_number=0, page_size=20):
        """
        Used to list all bounces for a specified message
        :param message_id: Mailup message id you need to analyze
        :return: {"IsPaginated":true,"Items":[
            {"Email":"clas@gimail.com","IdMessage":11775,"IdRecipient":1415989,"Type":"GB"}
        ], "PageNumber":0,"PageSize":10,"Skipped":0,"TotalElementsCount":1174}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.mailStatisticsEndpoint + "/Message/{}/List/Bounces".format(message_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_message_unsubscriptions(self, message_id, page_number=0, page_size=20):
        """
        Used to list all unsubscriptions from a specified messsage
        :param message_id: Mailup message id you need to analyze
        :return: {"IsPaginated":true,"Items":[
            {"Email":"manuel.f@example.com","IdMessage":11775,"IdRecipient":1419828},
        ],"PageNumber":0,"PageSize":8,"Skipped":0,"TotalElementsCount":179}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.mailStatisticsEndpoint + "/Message/{}/List/Unsubscriptions".format(message_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def read_message_clicks(self, message_id, page_number=0, page_size=20):
        """
        Used to list all clicks from a specified message
        :param message_id: Mailup message id you need to analyze
        :return: {"IsPaginated":true, "Items":[
            {"Count":283,"IdMessage":11826,"RecipientsClicks":[
                {"Count":2,"Email":"vladimir@example.com","IdRecipient":1419793},
                {"Count":3, "Email":"larry@example.com", "IdRecipient":1377520}
            ], "Url":""http:\/\/www.example.com\/responsability.htm""
        }],
        "PageNumber":0, "PageSize":2, "Skipped":0, "TotalElementsCount":300}
        """
        params = {"PageNumber":page_number, "PageSize": page_size}
        url = self.mailStatisticsEndpoint + "/Message/{}/List/UrlClickDetails".format(message_id)
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": "Bearer %s" % self.accessToken,
        }
        return self.call_method("GET", url, params=params, headers=headers)

    def call_method(self, method, url, data=None, params=None, headers=None, cookies=None, method_type="json"):
        try:
            req = getattr(requests, method.lower())(url, data=data, params=params, headers=headers, cookies=cookies)
            if req.status_code == 200:
                return req.json() if method_type == "json" else req.text
            elif req.status_code == 401:
                log.debug("Sleeping 5 seconds for code {} at {}: {}".format(req.status_code, req.url, req.text))
                time.sleep(5)
                self.refresh_token()
                if "Authorization" in headers and headers["Authorization"].startswith("Bearer"):
                    headers["Authorization"] = "Bearer %s" % self.accessToken
                self.call_method(method, url, data, params, headers, cookies, method_type)
            else:
                log.debug("Error code {} at {}: {}".format(req.status_code, req.url, req.text))
        except Exception, e:
            log.debug("Generic error in call_method: {}".format(e))
